/*
 * Copyright (c) 1997 Ian W. Dall <ian.dall@dsto.defence.gov.au>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Ian W. Dall.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <assert.h>
#include "main.h"
#include "menue.h"
#include "analyse_disk.h"
#include "scsi_op.h"
#include <time.h>

static unsigned int start_block, end_block;

static int start_block_validate(struct menue_entry *entryp) {
  if ((*entryp->u.data.value.Unsigned <= end_block)) {
    return 1;
  } else {
    message("Start block greater than end block.\n");
    return 0;
  }
}

static int end_block_validate(struct menue_entry *entryp) {
  if ((*entryp->u.data.value.Unsigned < start_block)) {
    message("End block less than start block.\n");
    return 0;
  } else if ((*entryp->u.data.value.Unsigned > current_disk_max_block)){
    message("End block greater than maximum block address(%u).\n",
	    current_disk_max_block);
    return 0;
  } else {
    return 1;
  }
}


static bool remap_disable = True;
static bool remap_on_hard = False;
static bool remap_on_soft = False;
static bool rewrite_disable = False;
static bool rewrite_on_soft = False;
static bool write_disable = True;

static int remap_on_error_hook (struct menue_entry *entryp);

static int rewrite_on_error_hook (struct menue_entry *entryp) {
  if(rewrite_on_soft) {
    remap_on_soft = False;
    remap_on_error_hook(0);
  }
  return 0;
}

static int remap_on_error_hook (struct menue_entry *entryp) {
  remap_disable = !(remap_on_hard || remap_on_soft);
  if(remap_on_soft)
    rewrite_on_soft = False;
  return 0;
}

static int write_disable_hook(struct menue_entry *entryp) {
  rewrite_disable = !write_disable;
  return 0;
}

struct ilist {
  struct ilist *next;
  int n;
};

struct ilist *write_patterns = 0;

static int _write_pattern_validate(char *p) {
  int ret;
  unsigned int tmp, n, nlist = 0;
  struct ilist *listp, *nextp, **last_next;
  /* Deallocate any list */
  for (listp = write_patterns; listp; listp = nextp) {
    nextp = listp->next;
    free(listp);
  }
  write_patterns = 0;

  last_next = &write_patterns;
  ret = sscanf(p, "%i%n", &tmp, &n);
  while (ret) {
    listp = (struct ilist *) malloc(sizeof(struct ilist));
    listp->n = tmp;
    listp->next = 0;
    *last_next = listp;
    last_next = &listp->next;
    nlist++;
    
    p += n;
    for(; *p == ' ' || *p == '\t' || *p == ','; p++);
    if( *p == '\0')
      break;
    ret = sscanf(p, "%i%n", &tmp, &n);
  }
  return nlist;
}

FILE *log;

static string log_file = "";
static int log_file_hook(struct menue_entry *entryp)
{
  FILE *temp = 0;
  if(strlen(*entryp->u.data.value.String)) {
    if(strcmp(log_file, *entryp->u.data.value.String) == 0)
      return 1;
    temp = fopen(*entryp->u.data.value.String, "w");
    if(temp == 0) {
      warning(*entryp->u.data.value.String);
      return 0;
    }
    if(log) {
      fclose(log);
    }
    log = temp;
    if(setvbuf(log, (char *)NULL, _IOLBF, 0) == EOF)
      warningx("Unable to set linebuf mode on stream %s",
	       *entryp->u.data.value.String);
  } else {
    if(log) {
      fclose(log);
      log = 0;
    }
  }
  return 1;
}

static int write_pattern_validate (struct menue_entry *entryp) {
  int n = _write_pattern_validate(*entryp->u.data.value.String);
  int new_len;
  char *p;
  struct ilist *listp;
  if (n == 0)
    return 0;
  new_len = n * 12 - 2;
  assert(entryp->u.data.malloced_value);
  free(*entryp->u.data.value.String);
  *entryp->u.data.value.String = (char *) malloc(new_len + 1);
  p = *entryp->u.data.value.String;
  if(!p)
    errorx("write_pattern_validate: Malloc failed.");
  for(listp = write_patterns; listp; listp = listp->next) {
    p += sprintf(p, "0x%08x%s", listp->n, listp->next? ", ": "");
    assert(p - *entryp->u.data.value.String <= new_len);
  }
  return 1;
}

/* Forward declaration */
static void
do_analysis(FILE *log, unsigned int start_block, unsigned int end_block,
	    unsigned int iterations, struct ilist *write_patterns,
	    bool remap_on_hard, bool remap_on_soft, bool rewrite_on_soft,
	    bool remap_confirm);


void analyse_disk(void) {
  enum {Exit = 1, Extended_Options, Write_Params, Doit};
  static unsigned int iterations = 1;
  static bool remap_confirm = True;
  static string write_pattern = "0xAAAAAAAA, 0x55555555, 0x00000000, 0xFFFFFFFF";
  static struct menue_entry entry_list[] = {
    ACTION("Extended Options.", 0, Extended_Options , 0),
    VAR_TOGGLE("Remap on Hard Error? ", 0, Bool, remap_on_hard,
	remap_on_error_hook),
    VAR_TOGGLE("Remap on Soft Error? ", 0, Bool, remap_on_soft,
	remap_on_error_hook),
    VAR_TOGGLE("Re-write on Soft error: ", &rewrite_disable, Bool,
	       rewrite_on_soft, rewrite_on_error_hook),
    VAR_TOGGLE("Query before remapping? ", &remap_disable, Bool, remap_confirm,
	       0),
    VAR_VALIDATE("Log File?:", 0, "%s", "%s", String, log_file, log_file_hook),
    VAR_RANGE("No of iterations: ", 0, "%u", "%i", Unsigned, iterations,
	      1, UINT_MAX),
    ACTION("Start Surface Analysis.", 0, Doit , 0),
    [9] = ACTION("Exit", 0, Exit, 0)
  };


  static struct menue analyse_menue = MENUE("Scan Disk", entry_list);

  static struct menue_entry extended_entries[] = {
    VAR_VALIDATE("Start Block No: ", 0, "%u", "%ui", Unsigned, start_block,
		 start_block_validate),
    VAR_VALIDATE("End Block No: ", 0, "%u", "%ui", Unsigned, end_block,
		 end_block_validate),
    VAR_TOGGLE("Read Only? ", 0, Bool, write_disable, write_disable_hook),
    VAR_VALIDATE("Write Pattern: ", &write_disable, "%s", "%s", String, write_pattern, write_pattern_validate),
    [9] = ACTION("Exit", 0, Exit, 0)
  };


  static struct menue extended_options_menue = MENUE("Scan Disk Extended Options", extended_entries);
  
  if(! current_disk) {
    message("No disk currently selected\n");
    return;  

  }
  start_block = 0;
  end_block = current_disk_max_block;
  if (!_write_pattern_validate(write_pattern)) {
    warningx("Bad write pattern");
  }

  while (1) {
    switch (do_menue(analyse_menue)) {
    case Extended_Options:
      if(do_menue(extended_options_menue) == EOF)
	clearerr(stdin);
      break;
    case Doit:
      do_analysis(log, start_block, end_block, iterations,
		  write_disable? 0: write_patterns,
		  remap_on_hard, remap_on_soft, rewrite_on_soft, remap_confirm);
      break;
    case EOF:
    default:
      clearerr(stdin);
      return;
    }
  }
}

#define WRITE_BUF (1 << 16)

static int rewrite(unsigned int blockno)
{
  int status;
  unsigned int *buf;
  buf = (unsigned int *) malloc(current_disk_block_size);
  (void) scsi_read(blockno, 1, buf, current_disk_block_size, &status);
  if(status != ST_OK && status != ST_SOFT)
    return status;
  (void) scsi_write(blockno, 1, buf, current_disk_block_size, &status);
  return status;
}


static int remap_copy(unsigned int blockno)
{
  int status;
  unsigned int *buf;
  buf = (unsigned int *) malloc(current_disk_block_size);
  (void) scsi_read(blockno, 1, buf, current_disk_block_size, &status);
  if(status != ST_OK && status != ST_SOFT)
    return status;
  scsi_remap(1, &blockno, &status);
  if(status != ST_OK)
    return status;
  (void) scsi_write(blockno, 1, buf, current_disk_block_size, &status);
  return status;
}

static void
do_analysis(FILE *log, unsigned int start_block, unsigned int end_block,
	    unsigned int iterations, struct ilist *write_patterns,
	    bool remap_on_hard, bool remap_on_soft, bool rewrite_on_soft,
	    bool remap_confirm)
{
  bool write = (write_patterns != 0);
  struct ilist *listp = write_patterns;
  unsigned int block, dblocks, nblocks, blockno;
  int status, i;
  unsigned int *buf;
  unsigned int bufsize;
  bool rewritten = False;
  unsigned int rewritten_blockno;
  dblocks = WRITE_BUF/current_disk_block_size;
  if (write) {
    /* Write listp->n from start block to end block inclusive*/
    int n;
    bufsize = (dblocks * current_disk_block_size);
    buf = (unsigned int *) malloc(bufsize);
    for (n = 0; n < bufsize/sizeof(unsigned int); n++) {
      buf[n] = listp->n;
    }
  }
  for(i = 0; i < iterations; i++) {
    time_t lasttime = 0, currenttime;
    for(block = start_block; block < end_block + 1;) {
      nblocks = end_block + 1 - block;
      if(nblocks > dblocks)
	nblocks = dblocks;
      if (write) {
	if((time(&currenttime) - lasttime)/2 != 0) {
	  lasttime = currenttime;
	  display_status("Write 0x%x and verify %d (%d%%)", buf[0],
			 i,
			 (block - start_block)*100/(end_block - start_block));
	}
	blockno = scsi_write_verify(log, block, nblocks, buf, bufsize, &status);
      } else {
	if((time(&currenttime) - lasttime)/2 != 0) {
	  lasttime = currenttime;
	  display_status("Verifying %d (%d%%)", i,
			 (block - start_block)*100/(end_block - start_block));
	}
	blockno = scsi_verify(log, block, nblocks, &status);
      }
      switch(status) {
      case ST_OK:
	block += nblocks;
	break;
      case ST_SOFT:
	if (!write && rewrite_on_soft &&
	    !(rewritten && (rewritten_blockno == blockno))) {
	  rewritten = True;
	  rewritten_blockno = blockno;
	  rewrite(blockno);
	  /* Verify that the re-written block works */
	  block = blockno;
	  break;
	}

	if (remap_on_soft && (!remap_confirm ||
			      prompt_y_or_n("Soft error. Remap block %d ", blockno))) {
	  if(write)
	    scsi_remap(1, &blockno, &status);
	  else
	    status = remap_copy(blockno);
	  if (status != ST_OK) {
	    warningx(0, "Failure to remap %d - skipping.", blockno);
	    block = blockno + 1;
	  }	else {
	    /* Verify that the re-mapped block works */
	    block = blockno;
	  }
	} else
	  block = blockno + 1;
	break;
      case ST_HARD:
	if (remap_on_hard && (!remap_confirm ||
			      prompt_y_or_n("Hard error. Remap block %d ", blockno))) {
	  scsi_remap(1, &blockno, &status);
	  if (status != ST_OK) {
	    warningx(0, "Failure to remap %d - skipping.", blockno);
	    block = blockno + 1;
	  }	else {
	    /* Verify that the re-mapped block works */
	    block = blockno;
	  }
	} else
	  block = blockno + 1;
	break;
      case ST_CONTINUE:
	/* Just log the error */
	block = blockno + 1;
	break;
      case ST_FATAL:
      default:
	warningx(0, "do_analyse: Unrecoverable error.");
	return;
      }
    }
    if(write) {
      listp = listp->next;
      if(! listp)
	listp = write_patterns;
    }
  }
  if(write)
    free(buf);
}
