/*
 * Copyright (c) 1997 Ian W. Dall <ian.dall@dsto.defence.gov.au>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Ian W. Dall.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/scsiio.h>
#include <unistd.h>
#include <sys/errno.h>

#include "version.h"
#include "menue.h"
#include "main.h"
#include "scsi_op.h"
#include "analyse_disk.h"
#include "format_disk.h"

char *current_disk = NULL;
int current_diskfd;
int current_lun;
unsigned current_disk_max_block;
unsigned current_disk_block_size;

extern	char *__progname;		/* from crt0.o */

#define USAGE " special ...\n"

static struct menue process_args(int argc, char **argv) {
  int i, j;
  struct menue menue;
  struct menue_entry *entry_list;
  entry_list = malloc(argc * sizeof(struct menue_entry));
  for(i = 0, j = 0; *argv; argv++, j++) {
    int fd;
    struct scsi_addr scsi_addr;
    if ((fd = open(*argv, O_RDONLY, 0600)) != -1) {
      if (ioctl(fd, SCIOCIDENTIFY, &scsi_addr) != -1) {
	char str[256];
	snprintf(str, sizeof(str),
		 "%s, bus: %d, target id: %d, target lun: %d",
		 *argv, scsi_addr.scbus, scsi_addr.target,
		 scsi_addr.lun);
	entry_list[i].text = strdup(str);
	entry_list[i].entry_type = Action;
	entry_list[i].disable = NULL;
	entry_list[i].u.action_code = ((j << 3) | (scsi_addr.lun & 7)) + 1;
	entry_list[i].hook = 0;
	i++;
      }
      close(fd);
    }
  }

  menue.title = "Disk Selection";
  menue.len = i;
  menue.list = realloc(entry_list, i * sizeof(struct menue_entry));
  return menue;
}


static void free_menue(struct menue menue) {
  int i;
  struct menue_entry *entryp = menue.list;
  if(menue.list) {
    for (i = 0; i < menue.len; i++, entryp++ ) {
      if(entryp->text)
	free(entryp->text);
    }
    free(menue.list);
  }
}

/* Only produce an error message if we don't succeed for *any* device */
static void probe_bus(int argc, char **argv) {
  int i;
  char *err = NULL;
  for(i = 0; *argv; argv++) {
    int fd;
    struct scsi_addr scsi_addr = {-1, -1, -1};
    if ((fd = open(*argv, O_RDWR, 0600)) != -1) {
      i++;
      if (ioctl(fd, SCIOCREPROBE, &scsi_addr) == -1) {
	err = strerror(errno);
      } else {
	close(fd);
	break;
      }
      close(fd);
    }
  }
  if (i == 0) {
    /* Not able to open any files */
    err = strerror(errno);
  }
  if (err)
    warningx("%s", err);
}

static void disk_select(struct menue disk_select_menue, char **argv) {
  int ret, diskn;
  int done = 0;
  do {
    ret = do_menue(disk_select_menue);
    if (ret == EOF) {
      clearerr(stdin);
      done = 1;
    } else {
      diskn = ret -1;
      if (current_disk)
	close(current_diskfd);
      current_disk = argv[diskn >> 3];
      current_lun = diskn & 7;
      if ((current_diskfd = open(current_disk, O_RDWR, 0600)) == -1) {
	error("opening %s", current_disk);
      }
      disk_info();
      if(prompt_y_or_n("Confirm selection. "))
	done = 1;
    }
  } while(!done);
}

static int top_level(int argc, char **argv) {
  enum { Exit = 1, Probe, Select, Info, Format, Analysis, Remap};
  static struct menue_entry entry_list[] = {
    ACTION("Select Disk", 0, Select, 0),
    ACTION("Info Disk", 0, Info, 0),
    ACTION("Probe Bus", 0, Probe, 0),
    ACTION("Format Selected Disk", 0, Format, 0),
    ACTION("Surface Analysis", 0, Analysis, 0),
    ACTION("Remap Blocks", 0, Remap, 0),
    [9] = ACTION("Exit", 0, Exit, 0)
  };
  static struct menue top_menue = MENUE("SCSI Disk Doctor", entry_list);

  struct menue disk_select_menue = process_args(argc, argv);

  disk_select(disk_select_menue, argv);

  while (1) {
    switch(do_menue(top_menue)) {
    case Select:
      disk_select(disk_select_menue, argv);
      break;
    case Probe:
      probe_bus(argc, argv);
      free_menue(disk_select_menue);
      disk_select_menue = process_args(argc, argv);
      break;
    case Info:
      disk_info();
      prompt_any("");
      break;
    case Format:
      format_disk();
      break;
    case Analysis:
      analyse_disk();
      break;
#ifdef UNIMPLIMENTED
    case Remap:
      remap_blocks();
      break;
#endif
    case Exit:
      return(0);
    default:
      return(-1);
    }
  }
}

int main(int argc, char **argv) {
  if (argc < 2) {
    warningx("%s", "No disks specified.");
    errorxx("SCSI Disk Doctor version %s\nUsage: %s <special> [<special> ...]",
	    version, __progname);
  }
  get_ttysize();
  return top_level(argc - 1, argv + 1);
}
