/*
 * Copyright (c) 1997 Ian W. Dall <ian.dall@dsto.defence.gov.au>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Ian W. Dall.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <err.h>
#include <ctype.h>
#include <string.h>
#include <sys/ioctl.h>

#include "menue.h"

/* Like fgets but swallows the rest of the line if it wont fit in
 * the buffer and discards trailing newline like gets.
 */
static int unacked_output = 0;

static struct ttysize ttysize;

void get_ttysize(void)
{
  if((ioctl(1, TIOCGWINSZ, &ttysize) == -1) &&
     (ioctl(0, TIOCGWINSZ, &ttysize) == -1)) {
    warning("get_ttysize");
    ttysize.ts_cols = 80;
    ttysize.ts_lines = 24;
  }
  if(ttysize.ts_cols == 0)
    ttysize.ts_cols = 80;
  if(ttysize.ts_lines == 0)
    ttysize.ts_lines = 24;
}

static char *fgetstr(char *arg, int n, FILE *stream, int do_realloc)
{
  int c;
  register char *tempp = arg;
  int bufsize = n;
  int delta = n;
  int i = 0;
  while ( (c = getc(stream)) != EOF && c != '\n')
    {
      if (i < bufsize)
	{
	  i++;
	  *tempp++ = (unsigned char) c;
	}
      else if (realloc) {
	arg = realloc(arg, bufsize + delta);
	tempp = arg + i;
	bufsize += delta;
	i++;
	*tempp++ = (unsigned char) c;
      }
    }
  *tempp = '\0';
  unacked_output = 0;
  return arg;
}

static int print_value(struct menue_entry *entryp) {
  int ret = 0;
#define CASE(type) \
  case type: \
  ret = printf(entryp->u.data.p_format, *entryp->u.data.value.type); \
  break

  switch(entryp->u.data.type) {
    CASE(Int);
    CASE(Unsigned);
  case String:
    ret = printf("%s", *entryp->u.data.value.String);
    break;
  case Bool:
    putchar((*entryp->u.data.value.Bool)? 'Y': 'N');
    break;
  case BoolComp:
    putchar((*entryp->u.data.value.Bool)? 'N': 'Y');
    break;
  }
#undef CASE
  return ret;
}

#ifdef UNUSED
static int sprint_value(char *buf, struct menue_entry *entryp) {
  int ret = 0;
#define CASE(type) \
  case type: \
  ret = sprintf(buf, entryp->u.data.p_format, *entryp->u.data.value.type); \
  break

  switch(entryp->u.data.type) {
    CASE(Int);
    CASE(Unsigned);
  case String:
    ret = sprintf(buf, "%s", *entryp->u.data.value.String);
    break;
  case Bool:
    *buf = (*entryp->u.data.value.Bool)? 'Y': 'N';
    break;
  case BoolComp:
    *buf = (*entryp->u.data.value.Bool)? 'N': 'Y';
    break;
  }
#undef CASE
  return ret;
}
#endif /* UNUSED */

static int scan_value(struct menue_entry *entryp) {
  int ret = 0;
  char buf[BUFSIZ];
  if (entryp->u.data.type == String) {
    if(entryp->u.data.malloced_value) {
      free(*entryp->u.data.value.String);
    }
    *entryp->u.data.value.String = (char *) malloc(BUFSIZ);
    if(*entryp->u.data.value.String == 0) {
      errorx("scan_value: Malloc failure.");
    }
    entryp->u.data.malloced_value = 1;
    *entryp->u.data.value.String = fgetstr(*entryp->u.data.value.String, sizeof(buf), stdin, 1);
    return 1;
  }


  (void) fgetstr(buf, sizeof(buf), stdin, 0);
#define CASE(type) \
  case type: \
  ret = sscanf(buf, entryp->u.data.s_format, entryp->u.data.value.type); \
  break

  switch(entryp->u.data.type) {
    CASE(Int);
    CASE(Unsigned);
  default:			/* Bool and BoolComp should never get here */
    ret = EOF;
  }
#undef CASE
  return ret;
}

#ifdef UNUSED
static int sscan_value(char * buf, struct menue_entry *entryp) {
  int ret = 0;
  if (entryp->u.data.value.type == String) {
    if(entryp->u.data.malloced_value) {
      free(*entryp->u.data.value);
    }
    *entryp->u.data.value.String = (char *) malloc(strlen(buf) + 1);
    if(*entryp->u.data.value.String == 0) {
      errorx("scan_value: Malloc failure.");
    }
    strcpy(*entryp->u.data.value.String, buf);
    return 1;
  }
#define CASE(type) \
  case type: \
  ret = sscanf(buf, entryp->u.data.s_format, entryp->u.data.value.type); \
  break

  switch(entryp->u.data.type) {
    CASE(Int);
    CASE(Unsigned);
  }
#undef CASE
  return ret;
}
#endif /* UNUSED */

static int default_validate(struct menue_entry *entryp) {
  int ret = 0;
#define CASE(type) \
  case type: \
  ret = (*entryp->u.data.value.type >= entryp->u.data.min.type) \
	 && (*entryp->u.data.value.type <= entryp->u.data.max.type); \
  if (!ret) { \
    printf("Value outside range ["); \
    printf(entryp->u.data.p_format, entryp->u.data.min.type); \
    printf(", "); \
    printf(entryp->u.data.p_format, entryp->u.data.max.type); \
    printf("].\n"); \
  } \
  break

  switch(entryp->u.data.type) {
    CASE(Int);
    CASE(Unsigned);
  default:			/* Bool and BoolComp should never get here */
    printf("Internal Error.\n");
    ret = 0;
  }
#undef CASE
  return ret;
}

static int poly_copy(struct menue_entry *from_entryp, struct menue_entry *to_entryp) {
  unsigned int size = 0;
  void *from_addr, *to_addr;

#define CASE(type)  case type: size = sizeof(((union poly*) 0)->type);\
   to_addr = (void *)to_entryp->u.data.value.type;\
   from_addr = (void *)from_entryp->u.data.value.type;\
   break

  switch(from_entryp->u.data.type) {
    CASE(Int);
    CASE(Unsigned);
    CASE(Bool);
    CASE(BoolComp);
  case String:
    size = strlen(*from_entryp->u.data.value.String) + 1;
    if(*to_entryp->u.data.value.String && to_entryp->u.data.malloced_value)
      free(*to_entryp->u.data.value.String);
    *to_entryp->u.data.value.String = (char *) malloc(size);
    if (*to_entryp->u.data.value.String == 0)
      errorx("Malloc failed!.");
    to_entryp->u.data.malloced_value = 1;
    strcpy(*to_entryp->u.data.value.String, *from_entryp->u.data.value.String);
    return 0;
  }
#undef CASE
   memcpy(to_addr, from_addr, size);
   return 0;
}

static int edit_value(struct menue_entry *entryp) {
  int ret;
  poly temp_poly;
  struct menue_entry temp_entry = *entryp;

#define CASE(type)  case type: \
   temp_entry.u.data.value.type = &temp_poly.type;\
   break

  switch(temp_entry.u.data.type) {
    CASE(Int);
    CASE(Unsigned);
    CASE(Bool);
    CASE(BoolComp);
  case String:
    temp_entry.u.data.value.String = &temp_poly.String;
    *temp_entry.u.data.value.String = 0;
    temp_entry.u.data.malloced_value = 0;
    break;
  }
#undef CASE


  do {
    printf("%s [", entryp->text);
      print_value(entryp);
    printf("]: ");
    ret = scan_value(&temp_entry);
    if (ret == EOF)
      goto done;
    if (ret != 1)
      continue;
    if (entryp->hook? entryp->hook(&temp_entry):
	default_validate(&temp_entry)) {
      poly_copy(&temp_entry, entryp);
      goto done;
    }
  } while (1);
done:
  if (temp_entry.u.data.type == String &&
      temp_entry.u.data.malloced_value)
    free(*temp_entry.u.data.value.String);
  return ret;
}

static int display_menue(struct menue menue, int out) {
  int i;
  int lines = 0;
  struct menue_entry *entryp = menue.list;
  if(unacked_output) {
    prompt_any("");
  }
  if(out) {
    int pad = ((int)ttysize.ts_cols - strlen(menue.title))/2;
    if (pad < 0)
      pad = 0;
    printf("%*s%s\n\n", pad, "", menue.title);
  }
  lines+=2;
  for (i = 0; i < menue.len; entryp++, i++) {
    switch((entryp->disable && *entryp->disable)? Space:
	   entryp->entry_type) {
    case Data:
      if (out) {
	printf("%d. %s = ", i + 1, entryp->text);
	print_value(entryp);
      }
      break;
    case Skip:
      continue;
      break;
    case Action:
      if (out)
	printf("%d. %s", i + 1, entryp->text);
      break;
    case Space:
      break;
    }
    if (out)
      putchar('\n');
    lines++;
  }
  return lines;
}


int do_menue(struct menue menue) {
  int ret, entry_no;
  char * msg = "";
  display_status("");
  while (1) {
    do {
      char buf[BUFSIZ];
      int lines = display_menue(menue, 0);
      int i;
      for(i = 0; i < (int) ttysize.ts_lines - (lines + 2); i++)
	putchar('\n');
      (void) display_menue(menue, 1);
      putchar('\n');
      printf("%sNo of Selection: ", msg);
      msg = "";
      (void) fgetstr(buf, sizeof(buf), stdin, 0);
      ret = sscanf(buf, "%d", &entry_no);
      if(ret == 1) {
	if((entry_no < 1) || (entry_no > menue.len + 1)) {
	  msg="Invalid Selection. ";
#ifdef DEBUG
	  printf("No out of range [%d, %d]\n", 1, menue.len);
#endif
	  ret = 0;
	}
      } else if (feof(stdin))
	return EOF;
    } while (ret != 1);
    {
      struct menue_entry *entryp;
      entryp = &menue.list[entry_no - 1];
      if(entryp->disable && *entryp->disable) {
	msg = "Invalid Selection. ";
      } else if(entryp->entry_type == Action) {
	if ((ret = entryp->hook? entryp->hook(entryp): entryp->u.action_code)) {
	  /* An action code of zero is always the null action */
	  return ret;
	}
      } else if (entryp->entry_type == Data) {
	if((entryp->u.data.type == Bool)
	   || (entryp->u.data.type == BoolComp)) {
	  *entryp->u.data.value.Bool = !*entryp->u.data.value.Bool;
	  if (entryp->hook && (ret = entryp->hook(entryp))) {
	    /* An action code of zero is always the null action */
	    return ret;
	  }
	} else {
	  ret = edit_value(entryp); /* edit_value calls hook */
	  if (ret == EOF)
	    clearerr(stdin);
	}
      } else {
	msg = "Invalid Selection. ";
      }
    }

  }
}

void error(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  verr(1, fmt, ap);
  va_end(ap);
}

void errorx(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  verrx(1, fmt, ap);
  va_end(ap);
}

void errorxx(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  putc('\n', stderr);
  exit(1);
}

void warning(const char *fmt, ...) {
  va_list ap;
  display_status("");
  va_start(ap, fmt);
  vwarn(fmt, ap);
  va_end(ap);
  unacked_output = 1;
}

void warningx(const char *fmt, ...) {
  va_list ap;
  display_status("");
  va_start(ap, fmt);
  vwarnx(fmt, ap);
  va_end(ap);
  unacked_output = 1;
}

void message(const char *fmt, ...) {
  va_list ap;
  display_status("");
  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  unacked_output = 1;
}

char *prompt(const char *fmt,...) {
  static char buf[BUFSIZ];
  va_list ap;
  display_status("");
  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  return fgetstr(buf, sizeof(buf), stdin, 0);
}

void prompt_any(const char *fmt,...) {
  static char buf[BUFSIZ];
  va_list ap;
  display_status("");
  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  printf("Return to continue:");
  fgetstr(buf, sizeof(buf), stdin, 0);
}

int prompt_y_or_n(const char *fmt,...) {
  char buf[BUFSIZ], *p;
  va_list ap;
  display_status("");
  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  printf("[Y/N]: ");
  fgetstr(buf, sizeof(buf), stdin, 0);
  for(p = buf; *p && isspace(*p); p++); /* Skip white space */
  if((*p != 'y') && (*p != 'Y'))
    return 0;
  p++;
  for(;*p && isspace(*p); p++); /* Skip white space */
  if (*p)
    return 0;
  else
    return 1;
}

void display_status(const char *fmt,...) {
  va_list ap;
  static struct {
    char *buf; int len;
  } current_status = {0, 0};
  int status_len, common_len, i;
  char *p1, *p2, buf[BUFSIZ];

  va_start(ap, fmt);
  status_len = vsnprintf(buf, BUFSIZ, fmt, ap);
  va_end(ap);
  if (current_status.buf) {
    if (current_status.len != status_len)
      current_status.buf = realloc(current_status.buf, status_len + 1);
  } else {
    current_status.buf = malloc(status_len + 1);
    current_status.len = 0;
    *current_status.buf = '\0';
  }
  for(common_len = 0, p1 = current_status.buf, p2 = buf;
      (*p1 == *p2) && (*p1 != '\0'); common_len++, p1++, p2++);

  for(i = 0; i < (current_status.len - common_len); i++)
    putchar('\b');
  printf("%*s", (current_status.len - common_len), "");
  for(i = 0; i < (current_status.len - common_len); i++)
    putchar('\b');
  fputs(buf + common_len, stdout);
  strcpy(current_status.buf, buf);
  current_status.len = status_len;
  fflush(stdout);
}
