/*
 * Copyright (c) 1997 Ian W. Dall <ian.dall@dsto.defence.gov.au>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Ian W. Dall.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/scsiio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <scsi/scsi_all.h>
#include <scsi/scsi_disk.h>

#include "main.h"
#include "menue.h"
#include "scsi_op.h"
#include "scsi_data.h"

#define DEFAULT_TIMEOUT 10000	/* 10 seconds in milliseconds */

/* Define a macro to do the boring bits of command setup. The data structures
 * may need further customization after this macro, depending on the command.
 * It is no longer necessary to put the lun in the cmd, but what about old
 * targets?
 */

#define CMD_SETUP(OP, LUN, REQ, CMD_SIZE, DATA_BLOCK, DATA_SIZE, FLAGS, TIMEOUT) \
  do {							    \
    memset(&(REQ), 0, sizeof(REQ));			    \
    (REQ).cmd[0] = (OP);				    \
    (REQ).cmd[1] = (LUN) << SCSI_CMD_LUN_SHIFT;		    \
    (REQ).cmdlen = (CMD_SIZE);			            \
    (REQ).databuf = (caddr_t) (DATA_BLOCK);		    \
    (REQ).datalen = (DATA_SIZE);			    \
    (REQ).senselen = SENSEBUFLEN;			    \
    (REQ).flags = (FLAGS);				    \
    (REQ).timeout = (TIMEOUT);		/* Ticks? */	    \
  } while (0)


static __inline void
_lto2b(u_int32_t val,u_int8_t *bytes)
{

	bytes[0] = (val >> 8) & 0xff;
	bytes[1] = val & 0xff;
}

static __inline void
_lto3b(u_int32_t val, u_int8_t *bytes)
{

	bytes[0] = (val >> 16) & 0xff;
	bytes[1] = (val >> 8) & 0xff;
	bytes[2] = val & 0xff;
}

static __inline void
_lto4b(u_int32_t val, u_int8_t *bytes)
{

	bytes[0] = (val >> 24) & 0xff;
	bytes[1] = (val >> 16) & 0xff;
	bytes[2] = (val >> 8) & 0xff;
	bytes[3] = val & 0xff;
}

static __inline u_int32_t
_2btol(u_int8_t *bytes)
{
	register u_int32_t rv;

	rv = (bytes[0] << 8) |
	     bytes[1];
	return (rv);
}

static __inline u_int32_t
_3btol(u_int8_t *bytes)
{
	register u_int32_t rv;

	rv = (bytes[0] << 16) |
	     (bytes[1] << 8) |
	     bytes[2];
	return (rv);
}

static __inline u_int32_t
_4btol(u_int8_t *bytes)
{
	register u_int32_t rv;

	rv = (bytes[0] << 24) |
	     (bytes[1] << 16) |
	     (bytes[2] << 8) |
	     bytes[3];
	return (rv);
}

static __inline void
_lto2l(u_int32_t val,u_int8_t * bytes)
{

	bytes[0] = val & 0xff;
	bytes[1] = (val >> 8) & 0xff;
}

static __inline void
_lto3l(u_int32_t val, u_int8_t *bytes)
{

	bytes[0] = val & 0xff;
	bytes[1] = (val >> 8) & 0xff;
	bytes[2] = (val >> 16) & 0xff;
}

static __inline void
_lto4l(u_int32_t val, u_int8_t *bytes)
{

	bytes[0] = val & 0xff;
	bytes[1] = (val >> 8) & 0xff;
	bytes[2] = (val >> 16) & 0xff;
	bytes[3] = (val >> 24) & 0xff;
}

static __inline u_int32_t
_2ltol(u_int8_t *bytes)
{
	register u_int32_t rv;

	rv = bytes[0] |
	     (bytes[1] << 8);
	return (rv);
}

static __inline u_int32_t
_3ltol(u_int8_t *bytes)
{
	register u_int32_t rv;

	rv = bytes[0] |
	     (bytes[1] << 8) |
	     (bytes[2] << 16);
	return (rv);
}

static __inline u_int32_t
_4ltol(u_int8_t *bytes)
{
	register u_int32_t rv;

	rv = bytes[0] |
	     (bytes[1] << 8) |
	     (bytes[2] << 16) |
	     (bytes[3] << 24);
	return (rv);
}

static int sc_cmd(scsireq_t *req) {
  int retries = 10;
  struct scsi_sense_data *sense = (struct scsi_sense_data *)(req->sense);
  while(retries--) {
    if ((ioctl(current_diskfd, SCIOCCOMMAND, (char *)req) != -1))
      break;
    if(errno == ENOMEM && retries) {
      warning("SCIOCCOMAND");
    }
    else {
      error("SCIOCCOMMAND");
      /* No reurn */
    }
  }
  switch (req->retsts) {
  case SCCMD_SENSE:
    warningx("Block %d: Sense key 0x%x, code 0x%x, qualifier 0x%x.",
	     _4btol(sense->info),
	     sense->flags & SSD_KEY,
	     sense->add_sense_code,
	     sense->add_sense_code_qual);

#if 0
    /* Print whole buffer for information and debugging */
    message("Sense bytes (%d): ", req->senselen_used);
    for (i=0;i < req->senselen_used;i++)
      message("%x%s", req->sense[i], (i < (req->senselen_used - 1))? ", ": "\n");
#endif
    return -1;
  case SCCMD_OK:
    break;
  case SCCMD_TIMEOUT:
    warningx("Command 0x%x timed out. %s", req->cmd[0], current_disk);
    return -1;
  case SCCMD_BUSY:
    warningx("Device Busy %s", current_disk);
    return -1;
  default:
    warningx("Unknown error during command 0x%x", req->cmd[0]);
    return -1;
  }
  return 0;
}

void disk_info(void) {
  scsireq_t req;

  if(! current_disk) {
    message("No disk currently selected\n");
    return;
  }
  message("Device %s currently selected\n\n", current_disk);

  {
    struct scsi_addr scsi_addr;

    if (ioctl(current_diskfd, SCIOCIDENTIFY, &scsi_addr) == -1) {
      warning("ioctl SCIOCIDENTIFY");
    } else {
      message("bus: %d, target id: %d, target lun: %d\n",
	      scsi_addr.scbus, scsi_addr.target,
	      scsi_addr.lun);
    }
  }

  /* Inquiry */
  {
    struct scsi_inquiry_data scsi_inquiry_data;
    struct scsi_inquiry *inquiry_cmd = (struct scsi_inquiry *) &req.cmd;
    CMD_SETUP(INQUIRY, current_lun, req, sizeof(struct scsi_inquiry),
	      &scsi_inquiry_data, sizeof(scsi_inquiry_data), SCCMD_READ, DEFAULT_TIMEOUT);
    memset(&scsi_inquiry_data, 0, sizeof(scsi_inquiry_data));
    inquiry_cmd->length = sizeof(scsi_inquiry_data);

    if (sc_cmd(&req) == 0) {
      message("%-10.10s '%8.8s'\n", "vendor:", scsi_inquiry_data.vendor);
      message("%-10.10s '%16.16s'\n", "product:",
	      scsi_inquiry_data.product);
      message("%-10.10s '%4.4s'\n", "revision:",
	      scsi_inquiry_data.revision);
      message("%-10.10s '%8.8s'\n", "extra:", scsi_inquiry_data.extra);
    }
  }


  /* Read Capacity */
  {
    struct scsi_read_cap_data  scsi_read_cap_data;

    CMD_SETUP(READ_CAPACITY, current_lun, req,
	      sizeof(struct scsi_read_capacity),
	      &scsi_read_cap_data, sizeof(scsi_read_cap_data), SCCMD_READ, DEFAULT_TIMEOUT);
    memset(&scsi_read_cap_data, 0, sizeof(scsi_read_cap_data));

    if (sc_cmd(&req) == 0) {
      unsigned long capacity, length;

      capacity = _4btol(scsi_read_cap_data.addr);
      length = _4btol(scsi_read_cap_data.length);
      message("Capacity: %lu blocks of %lu bytes.\n", capacity, length);
      current_disk_max_block = capacity;
      current_disk_block_size = length;
    }
  }
  /* Mode Sense */
}



unsigned int scsi_write_verify(FILE *log, unsigned int block, unsigned int nblocks,
			       unsigned int *buf, unsigned int buf_len,
			       int *status)
{
  scsireq_t req;
  unsigned int ret;

  /* Write Verify */
  {
    struct scsi_rw_big *write_verify_cmd = (struct scsi_rw_big *)&req.cmd;

    CMD_SETUP(WRITE_VERIFY, current_lun, req,
	      sizeof(struct scsi_rw_big),
	      buf, buf_len, SCCMD_WRITE, DEFAULT_TIMEOUT);

    _lto4b(block, write_verify_cmd->addr);
    _lto2b(nblocks, write_verify_cmd->length);

    if (sc_cmd(&req) == 0) {
      *status = ST_OK;
      ret = block + nblocks;
    } else {
      /* Interpret status */
      if(req.retsts == SCCMD_SENSE) {
	struct scsi_sense_data *sense = (struct scsi_sense_data *)&(req.sense);
	int key = sense->flags & SSD_KEY;
	ret = _4btol(sense->info);
	if(log)
	  fprintf(log,"Block %d: Sense key 0x%x, code 0x%x, qualifier 0x%x.\n",
	     ret,
	     sense->flags & SSD_KEY,
	     sense->add_sense_code,
	     sense->add_sense_code_qual);
	switch(key) {
	case SSD_KEY_RECOVERED:
	  *status = ST_SOFT;
	  break;
	case SSD_KEY_MEDIA:
	  *status = ST_HARD;
	  break;
	default:
	  *status = ST_FATAL;
	}
      } else {
	*status = ST_FATAL;
	ret = block;
      }
    }
  }
  return ret;
}

unsigned int scsi_verify(FILE *log, unsigned int block, unsigned int nblocks,
			 int *status)
{
  scsireq_t req;
  unsigned int ret;

  /* Verify */
  {
    struct scsi_rw_big *verify_cmd = (struct scsi_rw_big *)&req.cmd;
    int dummy;

    CMD_SETUP(VERIFY, current_lun, req,
	      sizeof(struct scsi_rw_big),
	      &dummy, 0, SCCMD_WRITE, DEFAULT_TIMEOUT);

    _lto4b(block, verify_cmd->addr);
    _lto2b(nblocks, verify_cmd->length);

    if (sc_cmd(&req) == 0) {
      *status = ST_OK;
      ret = block + nblocks;
    } else {
      /* Interpret status */
      if(req.retsts == SCCMD_SENSE) {
	struct scsi_sense_data *sense = (struct scsi_sense_data *)&(req.sense);
	int key = sense->flags & SSD_KEY;
	ret = _4btol(sense->info);
	if(log) {
	  fprintf(log,"Block %d: Sense key 0x%x, code 0x%x, qualifier 0x%x.\n",
	     ret,
	     sense->flags & SSD_KEY,
	     sense->add_sense_code,
	     sense->add_sense_code_qual);
	}
	switch(key) {
	case SSD_KEY_RECOVERED:
	  *status = ST_SOFT;
	  break;
	case SSD_KEY_MEDIA:
	  *status = ST_HARD;
	  break;
	case SSD_KEY_HARDWARE:
	  if(sense->add_sense_code == 0x15 &&
	     sense->add_sense_code_qual == 0x02)
	    *status = ST_HARD;
	  else
	    *status = ST_CONTINUE;
	  break;
	default:
	  *status = ST_FATAL;
	}
      } else {
	*status = ST_FATAL;
	ret = block;
      }
    }
  }
  return ret;
}

void scsi_remap(unsigned int nblocknos, unsigned int *blocknos, int *status)
{
  scsireq_t req;
  {
    struct scsi_reassign_blocks_data *data;
    int i;
    int datalen = (sizeof(struct scsi_reassign_blocks_data) +
		   sizeof(struct defect_descriptor) * (nblocknos - 1));
    data = (struct scsi_reassign_blocks_data *) malloc(datalen);
    if(!data)
      errorx("scsi_remap: malloc failure.");
    memset(data, 0, datalen);
    _lto2b(nblocknos * 4, data->length);
    for(i = 0; i < nblocknos; i++)
       _lto4b(blocknos[i], data->defect_descriptor[i].dlbaddr);
    CMD_SETUP(REASSIGN_BLOCKS, current_lun, req,
	      sizeof(struct scsi_reassign_blocks),
	      data, datalen, SCCMD_WRITE, DEFAULT_TIMEOUT);

    if (sc_cmd(&req) == 0) {
      *status = ST_OK;
    } else {
      /* Interpret status */
      if(req.retsts == SCCMD_SENSE) {
	struct scsi_sense_data *sense = (struct scsi_sense_data *)&(req.sense);
	int key = sense->flags & SSD_KEY;
	switch(key) {
	case SSD_KEY_RECOVERED:
	  *status = ST_SOFT;
	  break;
	case SSD_KEY_MEDIA:
	  *status = ST_HARD;
	  break;
	default:
	  *status = ST_FATAL;
	}
      } else {
	*status = ST_FATAL;
      }
    }
    free(data);
  }
  return;
}

unsigned int scsi_write(unsigned int block, unsigned int nblocks,
			       unsigned int *buf, unsigned int buf_len,
			       int *status)
{
  scsireq_t req;
  unsigned int ret;

  /* Write */
  {
    struct scsi_rw_big *write_cmd = (struct scsi_rw_big *)&req.cmd;

    CMD_SETUP(WRITE_BIG, current_lun, req,
	      sizeof(struct scsi_rw_big),
	      buf, buf_len, SCCMD_WRITE, DEFAULT_TIMEOUT);

    _lto4b(block, write_cmd->addr);
    _lto2b(nblocks, write_cmd->length);

    if (sc_cmd(&req) == 0) {
      *status = ST_OK;
      ret = block + nblocks;
    } else {
      /* Interpret status */
      if(req.retsts == SCCMD_SENSE) {
	struct scsi_sense_data *sense = (struct scsi_sense_data *)&(req.sense);
	int key = sense->flags & SSD_KEY;
	ret = _4btol(sense->info);
	switch(key) {
	case SSD_KEY_RECOVERED:
	  *status = ST_SOFT;
	  break;
	case SSD_KEY_MEDIA:
	  *status = ST_HARD;
	  break;
	default:
	  *status = ST_FATAL;
	}
      } else {
	*status = ST_FATAL;
	ret = block;
      }
    }
  }
  return ret;
}


unsigned int scsi_read(unsigned int block, unsigned int nblocks,
		       unsigned int *buf, unsigned int buf_len,
		       int *status)
{
  scsireq_t req;
  unsigned int ret;

  /* Read */
  {
    struct scsi_rw_big *read_cmd = (struct scsi_rw_big *)&req.cmd;

    CMD_SETUP(READ_BIG, current_lun, req,
	      sizeof(struct scsi_rw_big),
	      buf, buf_len, SCCMD_READ, DEFAULT_TIMEOUT);

    _lto4b(block, read_cmd->addr);
    _lto2b(nblocks, read_cmd->length);

    if (sc_cmd(&req) == 0) {
      *status = ST_OK;
      ret = block + nblocks;
    } else {
      /* Interpret status */
      if(req.retsts == SCCMD_SENSE) {
	struct scsi_sense_data *sense = (struct scsi_sense_data *)&(req.sense);
	int key = sense->flags & SSD_KEY;
	ret = _4btol(sense->info);
	switch(key) {
	case SSD_KEY_RECOVERED:
	  *status = ST_SOFT;
	  break;
	case SSD_KEY_MEDIA:
	  *status = ST_HARD;
	  break;
	default:
	  *status = ST_FATAL;
	}
      } else {
	*status = ST_FATAL;
	ret = block;
      }
    }
  }
  return ret;
}

void scsi_format(unsigned int nblocknos, unsigned int *blocknos, bool cmplist,
		 unsigned int interleave, unsigned int timeout, int *status)
{
  scsireq_t req;
  {
    struct format_unit_parameters *parameters;
    int i;
    int parameterlen = (sizeof(struct format_unit_parameters) +
		   sizeof(struct defect_descriptor) * (nblocknos - 1));
    parameters = (struct format_unit_parameters *) malloc(parameterlen);
    if(!parameters)
      errorx("scsi_format: malloc failure.");
    memset(parameters, 0, parameterlen);
    _lto2b(nblocknos * 4, parameters->header.length);
    for(i = 0; i < nblocknos; i++)
       _lto4b(blocknos[i], parameters->defect[i].dlbaddr);

    CMD_SETUP(FORMAT_UNIT, current_lun, req,
	      sizeof(struct scsi_reassign_blocks),
	      parameters, parameterlen, SCCMD_WRITE, timeout*1000);

    if (sc_cmd(&req) == 0) {
      *status = ST_OK;
    } else {
      /* Interpret status */
      if(req.retsts == SCCMD_SENSE) {
	struct scsi_sense_data *sense = (struct scsi_sense_data *)&(req.sense);
	int key = sense->flags & SSD_KEY;
	switch(key) {
	case SSD_KEY_RECOVERED:
	  *status = ST_SOFT;
	  break;
	case SSD_KEY_MEDIA:
	  *status = ST_HARD;
	  break;
	default:
	  *status = ST_FATAL;
	}
      } else {
	*status = ST_FATAL;
      }
    }
    free(parameters);
  }
  return;
}
